/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Side Effects Software Inc
 *	477 Richmond Street West
 *	Toronto, Ontario
 *	Canada   M5V 3E7
 *	416-504-9876
 *
 * NAME:	rsl_voplib.h ( VOP Library, RSL)
 *
 * COMMENTS:	This include file is used by VOPs to implement higher level
 *		shading functions.
 */

#ifndef __voplib__
#define __voplib__

#define M_SQRT1_2	0.7071067
#define M_PI		PI

#define VOP_MIN_FILTER_SIZE	1e-6
#define VOP_PERIOD_FADE		(1.0/8.0)

#define FILTERSTEP(e0, x0, x1, filter)	filterstep(e0,x0,x1,"filter",filter)
#define AREA(pp)			area(pp)
#define VOP_FADE_PERIODIC(v1, v0, average, fwidth)	\
	    mix((v1-v0)/fwidth, average, clamp(fwidth*VOP_PERIOD_FADE, 0, 1));

#define	lerp		mix

float
sinh(float v)
{
    return (exp(v)-exp(-v))/2;
}
float
cosh(float v)
{
    return (exp(v)+exp(-v))/2;
}
float
tanh(float v)
{
    return (exp(v)-exp(-v))/(exp(v)+exp(-v));
}

float
fit(float x, omin, omax, nmin, nmax)
{
    float	t;
    t = clamp((x - omin)/(omax - omin), 0, 1);
    return mix(nmin, nmax, t);
}

//
// Here are functions which exist in VEX, but have no equivalents in RSL
//
#define MAP_VOP_VECTOR_FUNCTIONS(fname,func)	\
float  vop_##fname(float v) { return func(v); } \
vector vop_##fname(vector v) { return vector(func(xcomp(v)), func(ycomp(v)), \
					func(zcomp(v))); } \
point  vop_##fname(point  v) { return point (func(xcomp(v)), func(ycomp(v)), \
					func(zcomp(v))); } \
normal vop_##fname(normal v) { return normal(func(xcomp(v)), func(ycomp(v)), \
					func(zcomp(v))); } \
color  vop_##fname(color v) { return color(func(comp(v,0)), func(comp(v,1)), \
					func(comp(v,2))); }

#define VOP_VECTOR_FUNCTIONS(fname)	\
	    MAP_VOP_VECTOR_FUNCTIONS(fname, fname)

VOP_VECTOR_FUNCTIONS(abs)
VOP_VECTOR_FUNCTIONS(sqrt)
VOP_VECTOR_FUNCTIONS(ceil)
VOP_VECTOR_FUNCTIONS(floor)
VOP_VECTOR_FUNCTIONS(radians)
VOP_VECTOR_FUNCTIONS(degrees)
VOP_VECTOR_FUNCTIONS(exp)
VOP_VECTOR_FUNCTIONS(acos)
VOP_VECTOR_FUNCTIONS(asin)
VOP_VECTOR_FUNCTIONS(atan)
VOP_VECTOR_FUNCTIONS(cosh)
VOP_VECTOR_FUNCTIONS(sinh)
VOP_VECTOR_FUNCTIONS(tanh)
VOP_VECTOR_FUNCTIONS(cos)
VOP_VECTOR_FUNCTIONS(sin)
VOP_VECTOR_FUNCTIONS(tan)
MAP_VOP_VECTOR_FUNCTIONS(rint, round)

float  frac(float x)  { return x - floor(x); } 
vector frac(vector x) { return x - vop_floor(x); }
point  frac(point x)  { return x - vop_floor(x); }
normal frac(normal x) { return x - vop_floor(x); }
color  frac(color x)  { return x - vop_floor(x); }

float  vop_avg(vector v)  { return (xcomp(v)+ycomp(v)+zcomp(v))/3.0; }
float  vop_avg(point v)  { return (xcomp(v)+ycomp(v)+zcomp(v))/3.0; }
float  vop_avg(normal v)  { return (xcomp(v)+ycomp(v)+zcomp(v))/3.0; }
float  vop_avg(color v)  { return (comp(v,0)+comp(v,1)+comp(v,2))/3.0; }
float  vop_mincomp(vector v)  { return min(xcomp(v), ycomp(v), zcomp(v)); }
float  vop_mincomp(point v)   { return min(xcomp(v), ycomp(v), zcomp(v)); }
float  vop_mincomp(normal v)  { return min(xcomp(v), ycomp(v), zcomp(v)); }
float  vop_mincomp(color v)   { return min(comp(v,0), comp(v,0), comp(v,0)); }
float  vop_maxcomp(vector v)  { return max(xcomp(v), ycomp(v), zcomp(v)); }
float  vop_maxcomp(point v)   { return max(xcomp(v), ycomp(v), zcomp(v)); }
float  vop_maxcomp(normal v)  { return max(xcomp(v), ycomp(v), zcomp(v)); }
float  vop_maxcomp(color v)   { return max(comp(v,0), comp(v,0), comp(v,0)); }

vector	vop_floattovec(float x, y, z)	{ return vector(x, y, z); }
point 	vop_floattovec(float x, y, z)	{ return point(x, y, z); }
normal	vop_floattovec(float x, y, z)	{ return normal(x, y, z); }
color	vop_floattovec(float x, y, z)	{ return color(x, y, z); }

matrix
vop_floattomatrix(float a00,a01,a02,a03, a10,a11,a12,a13,
			a20,a21,a22,a23, a30,a31,a32,a33)
{
    return matrix(a00,a01,a02,a03, a10,a11,a12,a13,
		  a20,a21,a22,a23, a30,a31,a32,a33);
}

#define vop_dot(a,b)	( (a) . (b) )
#define vop_cross(a,b)	( (a) ^ (b) )

void
vop_vectofloat(vector v; output float x, y, z)
{
    x = xcomp(v);
    y = ycomp(v);
    z = zcomp(v);
}

void
vop_vectofloat(point v; output float x, y, z)
{
    x = xcomp(v);
    y = ycomp(v);
    z = zcomp(v);
}

void
vop_vectofloat(normal v; output float x, y, z)
{
    x = xcomp(v);
    y = ycomp(v);
    z = zcomp(v);
}

void
vop_vectofloat(color v; output float x, y, z)
{
    x = comp(v, 0);
    y = comp(v, 0);
    z = comp(v, 0);
}

float	vop_mod(float x, y)		{ return mod(x,y); }
vector	vop_mod(vector v; float y)
    { return vector(mod(xcomp(v),y), mod(ycomp(v),y),  mod(zcomp(v),y)); }
point	vop_mod(point v; float y)
    { return point(mod(xcomp(v),y),  mod(ycomp(v),y),  mod(zcomp(v),y)); }
normal	vop_mod(normal v; float y)
    { return normal(mod(xcomp(v),y), mod(ycomp(v),y),  mod(zcomp(v),y)); }
color	vop_mod(color v; float y)
    { return color(mod(comp(v,0),y), mod(comp(v,0),y), mod(comp(v,0),y)); }

float	vop_pow(float x, y)		{ return pow(x,y); }
vector	vop_pow(vector v; float y)
    { return vector(pow(xcomp(v),y), pow(ycomp(v),y),  pow(zcomp(v),y)); }
point	vop_pow(point v; float y)
    { return point(pow(xcomp(v),y),  pow(ycomp(v),y),  pow(zcomp(v),y)); }
normal	vop_pow(normal v; float y)
    { return normal(pow(xcomp(v),y), pow(ycomp(v),y),  pow(zcomp(v),y)); }
color	vop_pow(color v; float y)
    { return color(pow(comp(v,0),y), pow(comp(v,0),y), pow(comp(v,0),y)); }

float	vop_fit(float v, s0, s1, d0, d1)
    { return fit(v, s0, s1, d0, d1); }
vector	vop_fit(vector v, s0, s1, d0, d1)
    { return vector(fit(xcomp(v), xcomp(s0), xcomp(s1), xcomp(d0), xcomp(d1)),
		    fit(ycomp(v), ycomp(s0), ycomp(s1), ycomp(d0), ycomp(d1)),
		    fit(zcomp(v), zcomp(s0), zcomp(s1), zcomp(d0), zcomp(d1)));
    }
point	vop_fit(point v, s0, s1, d0, d1)
    { return point(fit(xcomp(v), xcomp(s0), xcomp(s1), xcomp(d0), xcomp(d1)),
		    fit(ycomp(v), ycomp(s0), ycomp(s1), ycomp(d0), ycomp(d1)),
		    fit(zcomp(v), zcomp(s0), zcomp(s1), zcomp(d0), zcomp(d1)));
    }
normal	vop_fit(normal v, s0, s1, d0, d1)
    { return normal(fit(xcomp(v), xcomp(s0), xcomp(s1), xcomp(d0), xcomp(d1)),
		    fit(ycomp(v), ycomp(s0), ycomp(s1), ycomp(d0), ycomp(d1)),
		    fit(zcomp(v), zcomp(s0), zcomp(s1), zcomp(d0), zcomp(d1)));
    }
color	vop_fit(color v, s0, s1, d0, d1)
    { return color(
	   fit(comp(v,0), comp(s0,0), comp(s1,0), comp(d0,0), comp(d1,0)),
	   fit(comp(v,1), comp(s0,1), comp(s1,1), comp(d0,1), comp(d1,1)),
	   fit(comp(v,2), comp(s0,2), comp(s1,2), comp(d0,2), comp(d1,2)));
    }

float vop_smooth(float min, max, amount, rolloff)
{
    return smoothstep(min, max, amount);
}

#define VOP_GETCOMP(type) \
    float vop_getcomp(type v; float index) \
	{ return comp(color(v), index); }

VOP_GETCOMP(vector)
VOP_GETCOMP(point)
VOP_GETCOMP(normal)

float vop_getcomp(color c; float index)  
{ 
    return comp(c, index); 
}

float vop_getcomp(matrix m; float row, column)
{
    return comp(m, row, column);
}

#define VOP_SETCOMP(type) \
    type vop_setcomp(type in; float val, index) \
    { \
	color out = color(in); \
	setcomp(out, index, val); \
	return type(out); \
    }
VOP_SETCOMP(vector)
VOP_SETCOMP(point)
VOP_SETCOMP(normal)

color
vop_setcomp(color in; float val, index)
{
    color out = in;
    setcomp(out, index, val);
    return out;
}

matrix
vop_setmatcomp(matrix in; float fval, row, col)
{
    matrix out = in;
    setcomp(out, row, col, fval);
    return out;
}

void
vop_assign(output float fval11, fval12, fval13, fval14,
		        fval21, fval22, fval23, fval24,
		        fval31, fval32, fval33, fval34,
			fval41, fval42, fval43, fval44;
	   matrix mat)
{
    fval11 = comp(mat, 1, 1);
    fval12 = comp(mat, 1, 2);
    fval13 = comp(mat, 1, 3);
    fval14 = comp(mat, 1, 4);
    fval21 = comp(mat, 2, 1);
    fval22 = comp(mat, 2, 2);
    fval23 = comp(mat, 2, 3);
    fval24 = comp(mat, 2, 4);
    fval31 = comp(mat, 3, 1);
    fval32 = comp(mat, 3, 2);
    fval33 = comp(mat, 3, 3);
    fval34 = comp(mat, 3, 4);
    fval41 = comp(mat, 4, 1);
    fval42 = comp(mat, 4, 2);
    fval43 = comp(mat, 4, 3);
    fval44 = comp(mat, 4, 4);
}


void
vop_bindST(output float news; float ss, Sconnected;
	   output float newt; float tt, Tconnected)
{ news = (Sconnected != 0) ? ss : s; newt = (Tconnected != 0) ? tt : t;
}

void
vop_bindS(output float news; float ss, Sconnected)
{
    news = (Sconnected != 0) ? ss : s;
}

color
rgbtohsv(color rgb)
{
    return ctransform("rgb", "hsv", rgb);
}

color
hsvtorgb(color hsv)
{
    return ctransform("hsv", "rgb", hsv);
}

color
vop_colormix(color c1, c2; float bias, adjust)
{
    color	clr;
    if (adjust == 3)
	clr = spline("catrom", bias, c1, c1, c2, c2);
    else if (adjust == 2)
	clr = mix(c1, c2, float(smoothstep(0, 1, bias)));
    else if (adjust == 1)
	clr = mix(c1, c2, float(clamp(bias, 0, 1)));
    else
	clr = mix(c1, c2, bias);
    return clr;
}

float
vop_FilterWidth(float x)
{
    return max(abs(Du(x)*du)+abs(Dv(x)*dv), VOP_MIN_FILTER_SIZE);
}

float
vop_Pulse(float edge0, edge1, x, fwidth; string filter)
{
    float	x0, x1;

    x0 = x  - fwidth*.5;
    x1 = x0 + fwidth;
    return max(0, (min(x1, edge1)-max(x0, edge0))/fwidth);
}

float
vop_FilteredSin(float x, fwidth)
{
    float	x0, x1;

    x0 = x - fwidth * .5;
    x1 = x + fwidth;
    return (-M_SQRT1_2)*VOP_FADE_PERIODIC(cos(x1), cos(x0), 0, fwidth);
}

float
vop_FilteredSinD(float x)
{
    return vop_FilteredSin(x, vop_FilterWidth(x));
}

float
vop_FilteredCos(float x, fwidth)
{
    float	x0, x1;

    x0 = x - fwidth * .5;
    x1 = x + fwidth;
    return M_SQRT1_2*VOP_FADE_PERIODIC(sin(x1), sin(x0), 0, fwidth);
}

float
vop_FilteredCosD(float x)
{
    return vop_FilteredCos(x, vop_FilterWidth(x));
}

float
vop_PulseD(float edge0, edge1, x; string filter)
{
    return vop_Pulse(edge0, edge1, x, vop_FilterWidth(x), filter);
}

float
vop_PulseTrain(float edge0, x, fwidth)
{
    float	x0, x1;

    x0 = x  - fwidth*.5;
    x1 = x0 + fwidth;
    x0 = edge0*floor(x0) + min(edge0, frac(x0));
    x1 = edge0*floor(x1) + min(edge0, frac(x1));
    return VOP_FADE_PERIODIC(x1, x0, edge0, fwidth);
}

float
vop_PulseTrainD(float edge0, x)
{
    return vop_PulseTrain(edge0, x, vop_FilterWidth(x));
}

float
vop_RampTrain(float x, fwidth)
{
    float	x0, x1, f;
    x0 = x - fwidth*.5;
    x1 = x0 + fwidth;
    f = frac(x0); x0 = floor(x0) + f*f;
    f = frac(x1); x1 = floor(x1) + f*f;
    return .5*VOP_FADE_PERIODIC(x1, x0, .5, fwidth);
}

float
vop_RampTrainD(float x)
{
    return vop_RampTrain(x, vop_FilterWidth(x));
}

float
vop_IntegrateTent(float x)
{
    float	f;
    f = frac(x);
    if (f > .5) f = f*(2 - f) - .5;
    else	f = f*f;
    return .5*floor(x) + f;
}

float
vop_TentTrain(float x, fwidth)
{
    float	x0, x1;
    x0 = x - fwidth*.5;
    x1 = x0 + fwidth;
    x0 = vop_IntegrateTent(x0);
    x1 = vop_IntegrateTent(x1);
    return VOP_FADE_PERIODIC(x1, x0, .5, fwidth);
}

float
vop_TentTrainD(float x)
{
    return vop_RampTrain(x, vop_FilterWidth(x));
}

float
vop_DotStamp(float px, py, fwidth; string filter)
{
    float	d;
    d = px*px + py*py;
    return FILTERSTEP(1, d-fwidth, d+fwidth, filter);
}

float
vop_BumpStamp(float px, py, fwidth; string filter)
{
    float	d, v;
    float	x0, x1;
    d  = px*px + py*py;
    x0 = d  - fwidth*.5;
    x1 = x0 + fwidth;
    v  = 1-FILTERSTEP(1, d-fwidth, d+fwidth, filter);
    v *= 1-smoothstep(0.0, 1.0, d);
    return v;
}

float
vop_RoundCosStamp(float px, py, fwidth; float hexness, sides, power;
		    string filter)
{
    float	ss, tt;
    float	x0, x1;

    ss = atan(py, px);
    tt = px*px + py*py + hexness*pow(abs(1-cos(ss*(sides))), power);
    x0 = tt - fwidth*.5;
    x1 = x0 + fwidth;
    return FILTERSTEP(1, x0, x1, filter);
}

float
vop_RoundSinStamp(float px, py, fwidth; float hexness, sides, power;
		    string filter)
{
    float	ss, tt;
    float	x0, x1;

    ss = atan(py, px) - M_PI;
    tt = px*px + py*py + hexness*pow(abs(1-sin(ss*(sides))), power);
    x0 = tt - fwidth*.5;
    x1 = x0 + fwidth;
    return FILTERSTEP(1, x0, x1, filter);
}


float
vop_RingStamp(float px, py, iradius, oradius, fwidth; string filter)
{
    float	d;
    d = px*px + py*py;
    return vop_Pulse(iradius, oradius, d, fwidth, filter);
}

float
vop_BoxStamp(float px, py, fu, fv; string filter)
{
    float	dx;
    dx  = vop_Pulse(-1, 1, px, fu, filter);
    dx *= vop_Pulse(-1, 1, py, fv, filter);
    return dx;
}

//
// Tile generation
//  Input:	u, v	- The position
//		fx, fy	- Frequency
//		ox, oy	- Offset
//		jitter	- Jitter for each row
//  Return:	id	- Random ID for the tile
//  Output:	u, v,	- The coordinates within the tile
//		du, dv	- Derivative information
//  Unchanged:	fx, fy, ox, oy, jitter
//		
float
vop_TileGen(output float u, v; float fx, fy, ox, oy, stagger, jitter)
{
    float	row;

    v = v*fy - oy;
    row = floor(v);
    u = u*fx - ox - stagger*row + jitter*cellnoise(row);
    return floor(u) + row*1984;
}

float
vop_HexTileGen(output float u, v; float fx, fy, ox, oy)
{
    float	left, right;
    float	row, col;

    v   = v*fy - oy;
    row = floor(v);
    v  = frac(v);

    u = u*fx + ox;
    if (mod(floor(row), 2) == 1)
	u += .5;

    col = floor(u);
    u = frac(u);

    if (v > .5)
    {
	right = v - .5;
	left  = 1.5 - v;
	if (u > left)
	{
	    if (mod(floor(row), 2) == 0) col += 1;
	    row += 1;
	    v -= 1;
	    u = fit(u, left, 1, 0, .5);
	}
	else if (u < right)
	{
	    if (mod(floor(row), 2) == 1) col -= 1;
	    row += 1;
	    v -= 1;
	    u = fit(u, 0, right, 0.5, 1);
	}
	else u = fit(u, right, left, 0, 1);
    }
    v = (v*2+1)/3;
    return row * 938 + col;
}

//
// Stamping simple shape patterns.  Some patterns take auxilliary data.  These
// are:
//	ring	- aux data specifies the ring width (0 to 1)
//	hex	- aux specifies the "hexness" (0 to 1)
//
float
vop_StampPattern(float  layers;		// Number of layers
		 float dotsize;		// Base dotsize
		 float dj;		// Dotsize jitter
		 float softness;	// Softness
		 float px, jx;		// X position and jitter
		 float py, jy;		// Y position and jitter
		 float aux, jaux;	// Auxilliary data
		 string spottype;	// Type of spot
		 string filter;		// Filter type
		 output float floatseed;	// Return a random seed
	 )
{
    float	fwidth;
    float	lrandom;
    float	cx, cy, loff, dsize;
    float	fx, fy;
    float	djx, djy;
    float	du, dv;
    float	result, amount;
    float	irad, idot;
    point	pp;
    float	i;

    du = vop_FilterWidth(px) * softness;
    dv = vop_FilterWidth(py) * softness;
    fwidth = max(du, dv);
    result = 0;
    loff = 0;
    for (i = 0; i < layers; i += 1)
    {
	lrandom = cellnoise(i);
	pp = point(floor(px+loff), floor(py+loff),
		    1000.0*lrandom+500) + point(.5, .5, .5);
	floatseed = cellnoise(pp);

	pp = vector(cellnoise(pp)) - .5;
	dsize = dotsize * (1-dj*floatseed);
	idot = 1/dsize;

	djy  = idot*(1 - dsize);
	djx  = (2*clamp(jx, 0, 1)) * djy;
	djy *= (2*clamp(jy, 0, 1));

	cx = xcomp(pp) * djx;
	cy = ycomp(pp) * djy;
	fx = cx - 2*(frac(px+loff) - .5)*idot;
	fy = cy - 2*(frac(py+loff) - .5)*idot;

	if (spottype == "ring")
	{
	    irad = 1-clamp(aux + (zcomp(pp) - .5)*jaux, 0, 1);
	    amount = vop_RingStamp(fx, fy, irad, 1, fwidth, filter);
	}
	else if (spottype == "box")
	{
	    amount = vop_BoxStamp(fx, fy, du, dv, filter);
	}
	else if (spottype == "hex")
	{
	    amount = 1-vop_RoundCosStamp(fx, fy, fwidth, aux, 6, jaux, filter);
	}
	else if (spottype == "star")
	{
	    amount = 1-vop_RoundSinStamp(fx, fy, fwidth, aux, 5, jaux, filter);
	}
	else if (spottype == "bump")
	{
	    amount = vop_BumpStamp(fx, fy, fwidth, filter);
	}
	else
	{
	    amount = 1-vop_DotStamp(fx, fy, fwidth, filter);
	}
	result = (1-amount)*result + amount;
	loff += 0.5;
    }
    floatseed = (floatseed - 0.5)*32000;
    return result;
}

float
vop_RipplePattern(float x, y, decay, toff)
{
    float	d, v;

    d = (sqrt(x*x + y*y) - toff);
    v = sin(d) * exp(d*(-decay));
    return v;
}

#define VOP_FBMNOISE_FUNC(type)	\
	    float	scale, plimit, blend; \
	    float	octaves; \
	    plimit = 2*blur; \
	    nval = 0; octaves = 0; scale = 1; \
	    while (scale > plimit && octaves < maxoctaves) { \
		nval += scale *(type(noise(pp)) - 0.5); \
		scale *= rough; \
		pp += pp; \
		octaves += 1; \
	    } \
	    if (scale > blur) { \
		blend = scale * clamp(scale/blur - 1, 0, 1); \
		nval += blend*(type(noise(pp)) - 0.5); \
	    }

float
vop_fbmNoise(float pos; float rough; float maxoctaves)
{
    float	blur = vop_FilterWidth(pos);
    float	nval, pp = pos;
    VOP_FBMNOISE_FUNC( float )
    return nval;
}

float
vop_fbmNoise(point pos; float rough; float maxoctaves)
{
    float	blur = AREA(pos);
    float	nval;
    point	pp = pos;
    VOP_FBMNOISE_FUNC( float )
    return nval;
}

point
vop_fbmNoise(float pos; float rough; float maxoctaves)
{
    float	blur = vop_FilterWidth(pos);
    point	nval, pp = pos;
    VOP_FBMNOISE_FUNC( point )
    return nval;
}

point
vop_fbmNoise(point pos; float rough; float maxoctaves)
{
    float	blur = AREA(pos);
    point	nval, pp = pos;
    VOP_FBMNOISE_FUNC( point )
    return nval;
}

#define VOP_PERLINNOISE_FUNC(type)	\
	    vector	pp	= pos; \
	    float	scale	= amp; \
	    float	i; \
	    nval = 0; \
	    for (i = 0; i < turb; i += 1) { \
		nval += scale * type(noise(pp)); \
		pp *= 2.0; scale *= rough; \
	    } \
	    nval = (type(pow(nval, atten)));


float
vop_perlinNoise(point pos; float turb; float amp, rough, atten)
{
    float nval;
    VOP_PERLINNOISE_FUNC(float)
    return nval;
}

point
vop_perlinNoise(point pos; float turb; float amp, rough, atten)
{
    vector nval;
    VOP_PERLINNOISE_FUNC(vector)
    return nval;
}

void
vop_displaceAlongNormal(point pp; normal nn;
			float amount, scale, sshear, tshear;
			float for_poly, obj_space;
			output point dP; output normal dN)
{
    normal	ng;
    if (obj_space != 0)
    {
	dN = ntransform("current", "object", nn);
	ng = ntransform("current", "object", Ng);
	dP =  transform("current", "object", pp);
    }
    else
    {
	dN = nn;
	ng = Ng;
	dP = pp;
    }
    dP += (scale * amount) * dN;
    dP += (sshear * amount) * normalize(Du(dP));
    dP += (tshear * amount) * normalize(Dv(dP));
    if (for_poly != 0)
    {
	normal	dn = normalize(dN) - normalize(ng);	// Delta of normal
	dN = calculatenormal(dP);
	dN = normalize(dN) + dn;
    }
    else
    {
	dN = calculatenormal(dP);
    }
    if (obj_space != 0)
    {
	// TODO: If the transform doesn't have a negative determinant, then the
	// following code should probably not have a negative sign in front.
	dN = -ntransform("object", "current", dN);
	dP =   transform("object", "current", dP);
    }
}

void
vop_vnoise(float pos, jitter; output float seed, dist1, dist2; 
	   output float near1, near2)
{
    float thiscell = floor(pos) + 0.5;

    dist1 = 1000;  // initialize to a large number
    dist2 = 1000;
    near1 = 0;
    near2 = 0;
    uniform float   i;

    for (i = -1; i <= 1; i += 1) 
    {
	float  testcell = thiscell + i;
	float  cv       = float cellnoise(testcell) - 0.5;
	float  ptpos    = testcell + jitter * cv;

	float dist = ptpos - pos;
	if (dist < dist1)
	{
	    seed = testcell;
	    dist2 = dist1;
	    near2 = near1;

	    dist1 = dist;
	    near1 = ptpos;
	}
	else if (dist < dist2)
	{
	    dist2 = dist;
	    near2 = ptpos;
	}
    }
    seed = (float cellnoise(seed)-.5)*1000;
}

void
vop_vnoise(float pos, jitter; output float seed, dist1, dist2; 
	   output float near1, near2; float period)
{
    float thiscell = floor(pos) + 0.5;
	
    dist1 = 1000;  // initialize to a large number
    dist2 = 1000;
    near1 = 0;
    near2 = 0;
    uniform float   i;

    for (i = -1; i <= 1; i += 1) 
    {
	float  testcell = vop_mod(thiscell + i, period);
	float  cv       = float cellnoise(testcell) - 0.5;
	float  ptpos    = testcell + jitter * cv;

	float dist = ptpos - pos;
	if (dist < dist1)
	{
	    seed = testcell;
	    dist2 = dist1;
	    near2 = near1;

	    dist1 = dist;
	    near1 = ptpos;
	}
	else if (dist < dist2)
	{
	    dist2 = dist;
	    near2 = ptpos;
	}
    }
    seed = (float cellnoise(seed)-.5)*1000;
}

void
vop_vnoise_compute_dist(point pos, testcell, jitter; 
			output float dist1, dist2; output point near1, near2)
{
    vector cv       = vector cellnoise(testcell) - 0.5;
    point  ptpos    = testcell + vector(
			xcomp(jitter)*xcomp(cv),
			ycomp(jitter)*ycomp(cv),
			zcomp(jitter)*zcomp(cv));
    vector offset = ptpos - pos;
    float dist = offset . offset; // dist squared
    if (dist < dist1)
    {
	dist2 = dist1;
	near2 = near1;

	dist1 = dist;
	near1 = ptpos;
    }
    else if (dist < dist2)
    {
	dist2 = dist;
	near2 = ptpos;
    }
}

void
vop_vnoise(point pos, jitter; output float seed, dist1, dist2; 
	   output point near1, near2)
{
    point thiscell = point(floor(xcomp(pos)) + 0.5,
			   floor(ycomp(pos)) + 0.5,
			   floor(zcomp(pos)) + 0.5);
    dist1 = 1000;  // initialize to a large number
    dist2 = 1000;
    near1 = point(0, 0, 0);
    near2 = point(0, 0, 0);
    uniform float   i, j, k;

    for (i = -1; i <= 1; i += 1) 
    {
	for (j = -1; j <= 1; j += 1) 
	{
	    for (k = -1; k <= 1; k += 1)
	    {
		vop_vnoise_compute_dist(pos, thiscell + vector(i, j, k),
					jitter, dist1, dist2, near1, near2);
	    }
	}
    }

    seed = (float cellnoise(near1) - 0.5)*1000;
    dist1 = sqrt(dist1);
    dist2 = sqrt(dist2);
}

void
vop_vnoise(point pos, jitter; output float seed, dist1, dist2; 
	   output point near1, near2; point period)
{
    point thiscell = point(floor(xcomp(pos)) + 0.5,
			   floor(ycomp(pos)) + 0.5,
			   floor(zcomp(pos)) + 0.5);
    dist1 = 1000;  // initialize to a large number
    dist2 = 1000;
    near1 = point(0, 0, 0);
    near2 = point(0, 0, 0);
    uniform float   i, j, k;

    for (i = -1; i <= 1; i += 1) 
    {
	for (j = -1; j <= 1; j += 1) 
	{
	    for (k = -1; k <= 1; k += 1)
	    {
		point testcell = thiscell + vector(i, j, k);

		testcell = point(
			     vop_mod(xcomp(testcell), xcomp(period)),
			     vop_mod(ycomp(testcell), ycomp(period)),
			     vop_mod(zcomp(testcell), zcomp(period)) 
			   );

		vop_vnoise_compute_dist(pos, testcell, jitter, 
					dist1, dist2, near1, near2);
	    }
	}
    }

    seed = (float cellnoise(near1) - 0.5)*1000;
    dist1 = sqrt(dist1);
    dist2 = sqrt(dist2);
}

void
vop_vnoise(vector vpos; point jitter; output float seed, dist1, dist2; 
	   output point near1, near2; point period)
{
    vop_vnoise(point(vpos), jitter, seed, dist1, dist2, near1, near2, period);
}

void
vop_vnoise(vector vpos; point jitter; output float seed, dist1, dist2; 
	   output point near1, near2)
{
    vop_vnoise(point(vpos), jitter, seed, dist1, dist2, near1, near2);
}

#define VOP_VNOISE_VECTOR(type) \
void vop_vnoise(type vpos, jitter; output float seed, dist1, dist2; \
	        output type near1, near2) \
{ \
    point pnear1; \
    point pnear2; \
    vop_vnoise(point(vpos), point(jitter), seed, dist1, dist2, pnear1, pnear2);\
    near1 = type(pnear1); \
    near2 = type(pnear2); \
} \
void vop_vnoise(type vpos, jitter; output float seed, dist1, dist2; \
	        output type near1, near2; type period) \
{ \
    point pnear1; \
    point pnear2; \
    vop_vnoise(point(vpos), point(jitter), seed, dist1, dist2, pnear1, pnear2, \
	       point(period));\
    near1 = type(pnear1); \
    near2 = type(pnear2); \
}

VOP_VNOISE_VECTOR(vector)
VOP_VNOISE_VECTOR(color)
VOP_VNOISE_VECTOR(normal)

#define vop_frontface(nn, ii)	faceforward(nn, ii)

color
vop_anisotropic(normal nn; vector V; float urough, vrough)
{
    vector    LL;              // Normalized light vector
    vector    H;               // Half angle vector
    color    lclr;            // Light color
    vector    tanU, tanV;
    float     rz, cos_r, cos_i; // Reflected and incident angles

    float     nml_term;
    float     uval, vval;

    extern    vector dPdu;
    extern    vector dPdv;

    nml_term = 4.0 * M_PI * urough*vrough;

    cos_r = nn . V;
    lclr = 0;
    if (cos_r > 0.0)
    {
	tanU = normalize(dPdu) / urough;
	tanV = normalize(dPdv) / vrough;

        // Loop through all lights which contribute specular
        illuminance (P, nn, M_PI/2)
        {
	    float nonspec = 0;
	    lightsource("__nonspecular", nonspec);

	    LL = normalize(L);
	    cos_i = LL . nn;
            if (cos_i > 0.0)
            {
		 H = normalize (V + LL);
		 uval = tanU . H;
		 vval = tanV . H;
		 rz = cos_i*exp(-2.*(uval*uval + vval*vval) / 
				    (1.0 + H . nn)); 
		 rz /= nml_term * sqrt(cos_i*cos_r); 
		 lclr += (1 - nonspec) * Cl * rz; 
	    }
	}
    }
    else lclr = 0;
    return lclr;
}

color
vop_lighting(string lmodel; 
	     normal nf; vector ii; color amb, diff, spec; 
	     float urough, vrough)
{
    color clr;

    if (lmodel == "constant")
    {
	clr = diff;
    }
    else if (lmodel == "lambert")
    {
	clr = amb * ambient() + diff * diffuse(nf);
    }
    else if (lmodel == "headlight")
    {
	clr = diff * clamp((nf . -ii), 0, 1);
    }
    else if (lmodel == "oren")
    {
	// TODO: oren not yet implemented for rsl
	clr = amb * ambient() + diff * diffuse(nf);
    }
    else if (lmodel == "phong")
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * phong(nf, -ii, 1.0/urough);
    }
    else if (lmodel == "blinn")
    {
	// TODO: blinn not yet implemented for rsl
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * phong(nf, -ii, 1.0/urough);
    }
    else if (lmodel == "anisotropic")
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * vop_anisotropic(nf, -ii, urough, vrough);
    }
    else
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * specular(nf, -ii, urough);
    }
    return clr;
}

#endif
