/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Side Effects Software Inc
 *	477 Richmond Street West
 *	Toronto, Ontario
 *	Canada   M5V 3E7
 *	416-504-9876
 *
 * NAME:	voplib.h ( VOP Library Functions, VEX )
 *
 * COMMENTS:	Support functions for VOPS.  Most of these library functions
 *		are shading context only.
 */

#ifndef __voplib__
#define __voplib__

#if !defined(VOP_SHADING) && !defined(VOP_OP)
    #define VOP_OP
#endif

#define VOP_MIN_FILTER_SIZE	1e-6
#define VOP_PERIOD_FADE		(1.0/8.0)

#if defined(VOP_SHADING)
    #define FILTERSTEP(e0, x0, x1, filter) filterstep(e0,x0,x1,"filter",filter)
    #define AREA(pp)	area(pp)
#else
    #define FILTERSTEP(e0, x0, x1, filter) clamp(fit(e0, x0, x1, 0, 1), 0, 1)
    #define AREA(pp)		VOP_MIN_FILTER_SIZE
#endif

#define VOP_FADE_PERIODIC(v1, v0, average, fwidth)	\
	    lerp((v1-v0)/fwidth, average, clamp(fwidth*VOP_PERIOD_FADE, 0, 1));

#include <math.h>
#include <shading.h>

#define vop_abs		abs
#define vop_sqrt	sqrt
#define vop_ceil	ceil
#define vop_floor	floor
#define vop_radians	radians
#define vop_degrees	degrees
#define vop_exp		exp
#define vop_frac	frac
#define vop_avg		avg
#define vop_mincomp	min
#define vop_maxcomp	max
#define vop_rint	rint
#define vop_acos	acos
#define vop_asin	asin
#define vop_atan	atan
#define vop_cosh	cosh
#define vop_sinh	sinh
#define vop_tanh	tanh
#define vop_cos		cos
#define vop_sin		sin
#define vop_tan		tan
#define vop_fit		fit
#define vop_smooth	smooth
#define vop_getcomp	getcomp
#define vop_assign	assign
#define vop_vnoise	vnoise
#define vop_frontface	frontface
#define vop_dot		dot
#define vop_cross	cross

#define vop_floattovec(x,y,z)	set(x,y,z)
#define vop_vectofloat(v, x,y,z)	assign(x,y,z, v)
#define vop_floattomatrix(a00,a01,a02,a03, a10,a11,a12,a13, \
			  a20,a21,a22,a23, a30,a31,a32,a33) \
		set(a00,a01,a02,a03, a10,a11,a12,a13, \
		    a20,a21,a22,a23, a30,a31,a32,a33)

#define vop_mod(x,y)		((x) % (y))
#define vop_pow(x,y)		pow(x, y)

void
vop_bindST(float news, ss; int isSConnected;
	   float newt, tt; int isTConnected)
{
#if defined(VOP_SOP) || defined(VOP_POP)
    news = isSConnected ? ss : P.x;
    newt = isTConnected ? tt : P.y;
#elif defined(VOP_COP) || defined(VOP_COP2)
    news = isSConnected ? ss : X;
    newt = isTConnected ? tt : T;
#else
    news = isSConnected ? ss : s;
    newt = isTConnected ? tt : t;
#endif
}

void
vop_bindS(float news, ss; int isSConnected)
{
#if defined(VOP_SOP) || defined(VOP_POP)
    news = isSConnected ? ss : P.x;
#elif defined(VOP_COP) || defined(VOP_COP2)
    news = isSConnected ? ss : X;
#else
    news = isSConnected ? ss : s;
#endif
}

vector
vop_colormix(vector c1, c2; float bias; int adjust)
{
    vector	clr;
    if (adjust == 3)
	clr = cspline(bias, c1, c1, c2, c2);
    else if (adjust == 2)
	clr = lerp(c1, c2, float(smooth(0, 1, bias)));
    else if (adjust == 1)
	clr = lerp(c1, c2, float(clamp(bias, 0, 1)));
    else
	clr = lerp(c1, c2, bias);
    return clr;
}

float
vop_FilterWidth(float x)
{
#if defined(VOP_SHADING) || defined(VOP_COP2)
    return max(abs(Du(x))+abs(Dv(x)), VOP_MIN_FILTER_SIZE);
#else
    // There's really no way to determine what the filter size should be...
    return 0.005;
#endif
}

float
vop_Pulse(float edge0, edge1, x, fwidth; string filter)
{
    float	x0, x1;

    x0 = x  - fwidth*.5;
    x1 = x0 + fwidth;
    return max(0, (min(x1, edge1)-max(x0, edge0))/fwidth);
}

float
vop_FilteredSin(float x, fwidth)
{
    float	x0, x1;

    x0 = x - fwidth * .5;
    x1 = x + fwidth;
    return (-M_SQRT1_2)*VOP_FADE_PERIODIC(cos(x1), cos(x0), 0, fwidth);
}

float
vop_FilteredSinD(float x)
{
    return vop_FilteredSin(x, vop_FilterWidth(x));
}

float
vop_FilteredCos(float x, fwidth)
{
    float	x0, x1;

    x0 = x - fwidth * .5;
    x1 = x + fwidth;
    return M_SQRT1_2*VOP_FADE_PERIODIC(sin(x1), sin(x0), 0, fwidth);
}

float
vop_FilteredCosD(float x)
{
    return vop_FilteredCos(x, vop_FilterWidth(x));
}

float
vop_PulseD(float edge0, edge1, x; string filter)
{
    return vop_Pulse(edge0, edge1, x, vop_FilterWidth(x), filter);
}

float
vop_PulseTrain(float edge0, x, fwidth)
{
    float	x0, x1;

    x0 = x  - fwidth*.5;
    x1 = x0 + fwidth;
    x0 = edge0*floor(x0) + min(edge0, frac(x0));
    x1 = edge0*floor(x1) + min(edge0, frac(x1));
    return VOP_FADE_PERIODIC(x1, x0, edge0, fwidth);
}

float
vop_PulseTrainD(float edge0, x)
{
    return vop_PulseTrain(edge0, x, vop_FilterWidth(x));
}

float
vop_RampTrain(float x, fwidth)
{
    float	x0, x1, f;
    x0 = x - fwidth*.5;
    x1 = x0 + fwidth;
    f = frac(x0); x0 = floor(x0) + f*f;
    f = frac(x1); x1 = floor(x1) + f*f;
    return .5*VOP_FADE_PERIODIC(x1, x0, .5, fwidth);
}

float
vop_RampTrainD(float x)
{
    return vop_RampTrain(x, vop_FilterWidth(x));
}

float
vop_IntegrateTent(float x)
{
    float	f;
    f = frac(x);
    if (f > .5) f = f*(2 - f) - .5;
    else	f = f*f;
    return .5*floor(x) + f;
}

float
vop_TentTrain(float x, fwidth)
{
    float	x0, x1;
    x0 = x - fwidth*.5;
    x1 = x0 + fwidth;
    x0 = vop_IntegrateTent(x0);
    x1 = vop_IntegrateTent(x1);
    return VOP_FADE_PERIODIC(x1, x0, .5, fwidth);
}

float
vop_TentTrainD(float x)
{
    return vop_RampTrain(x, vop_FilterWidth(x));
}

float
vop_DotStamp(float px, py, fwidth; string filter)
{
    float	d;
    d = px*px + py*py;
    return FILTERSTEP(1, d-fwidth, d+fwidth, filter);
}

float
vop_BumpStamp(float px, py, fwidth; string filter)
{
    float	d, v;
    float	x0, x1;
    d  = px*px + py*py;
    x0 = d  - fwidth*.5;
    x1 = x0 + fwidth;
    v  = 1-FILTERSTEP(1, d-fwidth, d+fwidth, filter);
    v *= 1-smooth(0.0, 1.0, d);
    return v;
}

float
vop_RoundCosStamp(float px, py, fwidth; float hexness, sides, power;
		    string filter)
{
    float	ss, tt;
    float	x0, x1;

    ss = atan(py, px);
    tt = px*px + py*py + hexness*pow(abs(1-cos(ss*(sides))), power);
    x0 = tt - fwidth*.5;
    x1 = x0 + fwidth;
    return FILTERSTEP(1, x0, x1, filter);
}

float
vop_RoundSinStamp(float px, py, fwidth; float hexness, sides, power;
		    string filter)
{
    float	ss, tt;
    float	x0, x1;

    ss = atan(py, px) - M_PI;
    tt = px*px + py*py + hexness*pow(abs(1-sin(ss*(sides))), power);
    x0 = tt - fwidth*.5;
    x1 = x0 + fwidth;
    return FILTERSTEP(1, x0, x1, filter);
}


float
vop_RingStamp(float px, py, iradius, oradius, fwidth; string filter)
{
    float	d;
    d = px*px + py*py;
    return vop_Pulse(iradius, oradius, d, fwidth, filter);
}

float
vop_BoxStamp(float px, py, fu, fv; string filter)
{
    float	dx;
    dx  = vop_Pulse(-1, 1, px, fu, filter);
    dx *= vop_Pulse(-1, 1, py, fv, filter);
    return dx;
}

//
// Tile generation
//  Input:	u, v	- The position
//		fx, fy	- Frequency
//		ox, oy	- Offset
//		jitter	- Jitter for each row
//  Return:	id	- Random ID for the tile
//  Output:	u, v,	- The coordinates within the tile
//		du, dv	- Derivative information
//  Unchanged:	fx, fy, ox, oy, jitter
//		
int
vop_TileGen(float u, v; float fx, fy, ox, oy, stagger, jitter)
{
    int		row;

    v = v*fy - oy;
    row = floor(v);
    u = u*fx - ox - stagger*row + jitter*random(row);
    return floor(u) + row*1984;
}

int
vop_HexTileGen(float u, v; float fx, fy, ox, oy)
{
    float	left, right;
    int		row, col;

    v   = v*fy - oy;
    row = floor(v);
    v  = frac(v);

    u = u*fx + ox;
    if (row & 1)
	u += .5;

    col = floor(u);
    u = frac(u);

    if (v > .5)
    {
	right = v - .5;
	left  = 1.5 - v;
	if (u > left)
	{
	    if (!(row & 1)) col++;
	    row++;
	    v -= 1;
	    u = fit(u, left, 1, 0, .5);
	}
	else if (u < right)
	{
	    if (row & 1) col--;
	    row++;
	    v -= 1;
	    u = fit(u, 0, right, 0.5, 1);
	}
	else u = fit(u, right, left, 0, 1);
    }
    v = (v*2+1)/3;
    return row * 938 + col;
}

//
// This function generates Worley noise with a border.  Parameters are:
//	Input:	ss = S coord
//		tt = T coord
//		jx = Jitter in S direction
//		jy = Jitter in T direction
//		bwidth = Border width
//		bsoft = Border softness
//	Output:
//		Return code = 0 in border area, 1 in cell interior
//		centerx = X Center of closest node
//		centery = Y Center of closest node
//		seed = A random seed for the closest node
float
vop_aaCell2D(float ss, tt, jx, jy, bwidth, bsoft, centerx, centery;
		float f1, f2; int seed)
{
    float	p2x, p2y, x0;
    float	blur;
    vector	vp1, vp2;

    vnoise(ss, tt, jx, jy, seed, f1, f2, centerx, centery, p2x, p2y);

    vp1 = set(centerx, centery, 0);
    vp2 = set(p2x, p2y, 0);
    x0 = (f2-f1)*(f1+f2) / max(distance(vp1, vp2), VOP_MIN_FILTER_SIZE);

    blur = max(vop_FilterWidth(ss), vop_FilterWidth(tt)) * (1 + bsoft);

    return FILTERSTEP(bwidth, x0-blur, x0+blur, "gauss");
}


//
// Stamping simple shape patterns.  Some patterns take auxilliary data.  These
// are:
//	ring	- aux data specifies the ring width (0 to 1)
//	hex	- aux specifies the "hexness" (0 to 1)
//
float
vop_StampPattern(int   layers;		// Number of layers
		 float dotsize;		// Base dotsize
		 float dj;		// Dotsize jitter
		 float softness;	// Softness
		 float px, jx;		// X position and jitter
		 float py, jy;		// Y position and jitter
		 float aux, jaux;	// Auxilliary data
		 string spottype;	// Type of spot
		 string filter;		// Filter type
		 float	floatseed;	// Return a random seed
	 )
{
    float	fwidth;
    float	lrandom;
    float	cx, cy, loff, dsize;
    float	fx, fy;
    float	djx, djy;
    float	du, dv;
    float	result, amount;
    float	irad, idot;
    vector	pp;
    int		i;

    du = vop_FilterWidth(px) * softness;
    dv = vop_FilterWidth(py) * softness;
    fwidth = max(du, dv);
    result = 0;
    loff = 0;
    for (i = 0; i < layers; i++, loff += .5)
    {
	lrandom = random(i);
	pp = set(floor(px+loff), floor(py+loff),
		    1000.0*lrandom+500) + {.5, .5, .5};
	floatseed = random(pp);

	pp = vector(random(pp)) - .5;
	dsize = dotsize * (1-dj*floatseed);
	idot = 1/dsize;

	djy  = idot*(1 - dsize);
	djx  = (2*clamp(jx, 0, 1)) * djy;
	djy *= (2*clamp(jy, 0, 1));

	cx = pp.x * djx;
	cy = pp.y * djy;
	fx = cx - 2*(frac(px+loff) - .5)*idot;
	fy = cy - 2*(frac(py+loff) - .5)*idot;

	if (spottype == "ring")
	{
	    irad = 1-clamp(aux + (pp.z - .5)*jaux, 0, 1);
	    amount = vop_RingStamp(fx, fy, irad, 1, fwidth, filter);
	}
	else if (spottype == "box")
	{
	    amount = vop_BoxStamp(fx, fy, du, dv, filter);
	}
	else if (spottype == "hex")
	{
	    amount = 1-vop_RoundCosStamp(fx, fy, fwidth, aux, 6, jaux, filter);
	}
	else if (spottype == "star")
	{
	    amount = 1-vop_RoundSinStamp(fx, fy, fwidth, aux, 5, jaux, filter);
	}
	else if (spottype == "bump")
	{
	    amount = vop_BumpStamp(fx, fy, fwidth, filter);
	}
	else
	{
	    amount = 1-vop_DotStamp(fx, fy, fwidth, filter);
	}
	result = (1-amount)*result + amount;
    }
    floatseed = (floatseed - 0.5)*32000;
    return result;
}

float
vop_RipplePattern(float x, y, decay, toff)
{
    float	d, v;

    d = (sqrt(x*x + y*y) - toff);
    v = sin(d) * exp(d*(-decay));
    return v;
}

#define VOP_FBMNOISE_FUNC(type)	\
	    float	scale, plimit, blend; \
	    int		octaves; \
	    plimit = 2*blur; \
	    nval = 0; octaves = 0; scale = 1; \
	    while (scale > plimit && octaves < maxoctaves) { \
		nval += scale *(type(noise(pp)) - 0.5); \
		scale *= rough; \
		pp += pp; \
		octaves++; \
	    } \
	    if (scale > blur) { \
		blend = scale * clamp(scale/blur - 1, 0, 1); \
		nval += blend*(type(noise(pp)) - 0.5); \
	    }

float
vop_fbmNoiseFF(float pos; float rough; int maxoctaves)
{
    float	blur = vop_FilterWidth(pos);
    float	nval, pp = pos;
    VOP_FBMNOISE_FUNC( float )
    return nval;
}

float
vop_fbmNoiseFV(vector pos; float rough; int maxoctaves)
{
    float	blur = AREA(pos);
    vector	nval, pp = pos;
    VOP_FBMNOISE_FUNC( float )
    return nval;
}

float
vop_fbmNoiseFP(vector4 pos; float rough; int maxoctaves)
{
    float	blur = AREA((vector)pos);
    vector4	nval, pp = pos;
    VOP_FBMNOISE_FUNC( float )
    return nval;
}

vector
vop_fbmNoiseVF(float pos; float rough; int maxoctaves)
{
    float	blur = vop_FilterWidth(pos);
    float	nval, pp = pos;
    VOP_FBMNOISE_FUNC( vector )
    return nval;
}

vector
vop_fbmNoiseVV(vector pos; float rough; int maxoctaves)
{
    float	blur = AREA(pos);
    vector	nval, pp = pos;
    VOP_FBMNOISE_FUNC( vector )
    return nval;
}

vector
vop_fbmNoiseVP(vector4 pos; float rough; int maxoctaves)
{
    float	blur = AREA((vector)pos);
    vector4	nval, pp = pos;
    VOP_FBMNOISE_FUNC( vector )
    return nval;
}

#define VOP_PERLINNOISE_FUNC(type)	\
	    vector	pp	= pos; \
	    float	scale	= amp; \
	    int		i; \
	    nval = 0; \
	    for (i = 0; i < turb; i++, pp *= 2.0, scale *= rough) \
		nval += scale * (type(noise(pp))); \
	    nval = (type(pow(nval, atten)));


float
vop_perlinNoiseVF(vector pos; int turb; float amp, rough, atten)
{
    float nval;
    VOP_PERLINNOISE_FUNC(float)
    return nval;
}

vector
vop_perlinNoiseVV(vector pos; int turb; float amp, rough, atten)
{
    vector nval;
    VOP_PERLINNOISE_FUNC(vector)
    return nval;
}

void
vop_displaceAlongNormal(vector pp, nn; float amount, scale, sshear, tshear;
			    int	 for_poly, obj_space;
			    vector dP, dN)
{
    vector ng;
#if defined(VOP_SHADING)
    if (obj_space)
    {
	dN = -wo_nspace(nn);
	ng = -wo_nspace(Ng);
	dP =  wo_space(pp);
    }
    else
    {
	dN = nn;
	ng = Ng;
	dP = pp;
    }
#else
    dN = nn;
    dP = pp;
#endif
    dP += (scale * amount)*dN;
#if defined(VOP_SHADING)
    dP += (sshear * amount) * normalize(Du(dP));
    dP += (tshear * amount) * normalize(Dv(dP));
    if (for_poly)
	 dN = computenormal(dP, dN, ng);
    else dN = computenormal(dP);

    if (obj_space)
    {
	dN = -ow_nspace(dN);
	dP =  ow_space(dP);
    }
#endif
}

vector
vop_setcomp(vector in; float fval; int part)
{
    vector out = in;
    setcomp(out, fval, part);
    return out;
}

matrix
vop_setmatcomp(matrix in; float fval; int row, col)
{
    matrix out = in;
    setcomp(out, fval, row, col);
    return out;
}


vector
vop_anisotropic(vector nn; vector V; float urough, vrough)
{
    vector    LL;              // Normalized light vector
    vector    H;               // Half angle vector
    vector    lclr;            // Light color
    vector    tanU, tanV;
    float     rz, cos_r, cos_i; // Reflected and incident angles

    float     nml_term;
    float     uval, vval;

    nml_term = 4.0 * M_PI * urough*vrough;

    cos_r = dot(nn, V);
    lclr = 0;
    if (cos_r > 0.0)
    {
	tanU = normalize(dPds) / urough;
	tanV = normalize(dPdt) / vrough;

        // Loop through all lights which contribute specular
        illuminance (P, nn, M_PI/2, LIGHT_SPECULAR)
        {
	    LL = normalize(L);
	    cos_i = dot(LL , nn);
            if (cos_i > 0.0)
            {
		 H = normalize (V + LL);
		 uval = dot(tanU, H);
		 vval = dot(tanV, H);
		 rz = cos_i*exp(-2.*(uval*uval + vval*vval) / 
				    (1.0 + dot(H, nn))); 
		 rz /= nml_term * sqrt(cos_i*cos_r); 
		 lclr += Cl * rz; 
	    }
	}
    }
    else lclr = 0;
    return lclr;
}

vector
vop_lighting(string lmodel; 
	     vector nf, ii, amb, diff, spec; 
	     float urough, vrough)
{
    vector clr;

    if (lmodel == "constant")
    {
	clr = diff;
    }
    else if (lmodel == "lambert")
    {
	clr = amb * ambient() + diff * diffuse(nf);
    }
    else if (lmodel == "headlight")
    {
	clr = diff * clamp(dot(nf, -ii), 0, 1);
    }
    else if (lmodel == "oren")
    {
	clr = amb * ambient() + diff * diffuse(nf, -ii, $urough);
    }
    else if (lmodel == "phong")
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * phong(nf, -ii, 1.0/urough);
    }
    else if (lmodel == "blinn")
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * blinn(nf, -ii, urough);
    }
    else if (lmodel == "anisotropic")
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * vop_anisotropic(nf, -ii, urough, vrough);
    }
    else
    {
	clr = amb * ambient() + diff * diffuse(nf) + 
	      spec * specular(nf, -ii, urough);
    }
    return clr;
}



#endif
