/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *      Side Effects
 *      477 Richmond Street West
 *      Toronto, Ontario
 *      Canada   M5V 3E7
 *      416-504-9876
 *
 * NAME:        IFD library (C++)
 *
 * COMMENTS:    Sample code to output a RIB stream.  This code is limited to
 *		only outputting NURBs and Polygons.  However, it's only
 *		intended as an example.
 */

#ifdef WIN32
   #error This HDK sample does not support Windows NT (it requires RenderMan, which is not currently available for Windows NT).
#endif

#include <sys/time.h>
#include </usr/local/prman/include/ri.h>
#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <UT/UT_DMatrix4.h>
#include <GB/GB_NUBBasis.h>
#include <GEO/GEO_PrimPoly.h>
#include <GEO/GEO_PrimNURBSurf.h>
#include <GU/GU_Detail.h>
#include <OP/OP_Director.h>
#include <IFD/IFD_Scene.h>
#include <IFD/IFD_Entry.h>
#include "IFD_SimpleRib.h"

//
// The renderName here should not contain any spaces.  It is used as
//	an identifier.
static const char	*renderName = "SimpleRib";

// The IFD scene will use the following token to get information specific to
// RenderMan (vs. mantra or other shaders).
static const char	*RenderManToken = "RIB";

IFD_SimpleRib::IFD_SimpleRib()
{
    myLightHandles = 0;
    myGeometryHandles = 0;
}

IFD_SimpleRib::~IFD_SimpleRib()
{
    delete [] myLightHandles;
    delete [] myGeometryHandles;
}

const char *
IFD_SimpleRib::renderer(void) const { return renderName; }

void
IFD_SimpleRib::getOutputFlags(IFD_OutputFlags &flags) const
{
    flags.renderImage		= 1;	// Generates an image
    flags.renderScript		= 1;	// Supports a script interface
    flags.renderCommand		= 0;	// We don't support rendering to a cmd

    // Now, here we have a suggested command string for Houdini
    flags.suggestedCommand	= "render";
}

int
IFD_SimpleRib::doSingleFrame(IFD_Scene &scene, IFD_RenderParms &rparms,
				float now)
{
    const char	*scriptname;
    // Unlike the custom render command, we can't do a good open here.  The RIB
    // output does all it's own file management, so we have to do some work
    // here.
    if (!rparms.getScriptName)
	return 0;		// User's got to fill out a script name

    scriptname = rparms.getScriptName(&scene, now);
    if (!scriptname || !*scriptname)
	return 0;		// User has to fill out a script name

    RiBegin((char *)scriptname);	// Start the RIP stream

    myObjectCount = scene.getNObjects(now);
    myGeometryCount = scene.getNGeometry(now);
    myLightCount  = scene.getNLights(now);
    myBinaryFlag  = rparms.myBinaryFlag;
    myStartTime = now;
    myEndTime = now;
    myLightHandles = new void *[myLightCount+1];
    myGeometryHandles = new void *[myGeometryCount+1];

    outputGeometry(scene);	// Generate geometry to instance
    writeHeader(scene);		// Start rendering the frame
    outputLights(scene);	// Output light information
    outputObjects(scene);	// Instance the geometry
    writeTrailer();		// Finish the frame

    RiEnd();		// We're all done

    return 1;
}

void
IFD_SimpleRib::writeHeader(IFD_Scene &scene)
{
    float		 zoom, dl, dr, dt, db;
    UT_Vector3		 ambient;
    IFD_CameraParms	 parms;
    UT_String		 str;
    UT_String		 outfile, outdev;
    UT_DMatrix4		 xform;
    UT_Matrix4		 fxform;	// Single precision transform
    int			 frame;
    float		 aspect_fix;

    // Find out where we're outputting to
    scene.getCameraParms(parms,   myStartTime, RenderManToken);
    scene.getCameraTransform(xform, myStartTime);
    scene.getOutputImage(outfile, myStartTime, RenderManToken);
    scene.getOutputDevice(outdev, myStartTime, RenderManToken);

    frame = (int)OPgetDirector()->getChannelManager()->getSample(myStartTime);

    RiFrameBegin(frame);
    RiDisplay((char *)outfile, (char *)outdev, "rgba", RI_NULL);
    RiFormat(parms.xres, parms.yres, 1.0);
    RiExposure(parms.gain, parms.gamma);

    // Now, there's some tricky math here to get the correct screen window for
    // RenderMan.
    aspect_fix = parms.aspect * (float)parms.xres / (float)parms.yres;
    zoom = 2.0F * parms.focal / parms.aperture;
    if (parms.projection == IFD_ORTHO)
    {
	zoom = 2.0F * parms.orthoZoom / zoom;
	dl = (parms.winx[0] - 0.5F) * zoom;
	dr = (parms.winx[1] - 0.5F) * zoom;
	db = (parms.winy[0] - 0.5F) * zoom;
	dt = (parms.winy[1] - 0.5F) * zoom;
	RiScreenWindow(dl, dr, db/aspect_fix, dt/aspect_fix);
	RiProjection("orthographic", RI_NULL);
    }
    else
    {
	dl = (parms.winx[0] - 0.5F) * 2.0F;
	dr = (parms.winx[1] - 0.5F) * 2.0F;
	db = (parms.winx[0] - 0.5F) * 2.0F;
	dt = (parms.winx[1] - 0.5F) * 2.0F;
	RiScreenWindow(dl, dr, db/aspect_fix, dt/aspect_fix);

	zoom = 2.0F * UTradToDeg(atan2f(1.0F, zoom));
	RiProjection("perspective", "fov", &zoom, RI_NULL);
    }
    RiClipping(parms.mynear, parms.myfar);
    RiPixelSamples(parms.samplex, parms.sampley);

    xform.invert();		// Since it's a camera, we have to invert
    xform.scale(1, 1, -1);	// Put into the correct space

    fxform = xform;
    RiTransform(fxform.matx);	// Get the raw 4x4 array for Ri

    RiWorldBegin();	// Start off the world.
}

void
IFD_SimpleRib::outputMatrix(const UT_DMatrix4 &xform, int change_space)
{
    float	matx[4][4];
    int		i, j;

    for (i = 0; i < 4; i++)
	for (j = 0; j < 4; j++)
	    matx[i][j] = xform(i, j);

    if (change_space)
    {
	matx[0][3] = -matx[0][3];
	matx[1][3] = -matx[1][3];
	matx[2][3] = -matx[2][3];
	matx[3][3] = -matx[3][3];
    }
    RiTransform(matx);
}

void
IFD_SimpleRib::outputGeometry(IFD_Scene &scene)
{
    int			 i;
    IFD_GeometryParms	 parms;
    const GU_Detail	*gdp;

    for (i = 0; i < myGeometryCount; i++)
    {
	scene.getGeometryParms(i, parms, myStartTime, RenderManToken);
	gdp = scene.getGeometry(i, myStartTime);

	myGeometryHandles[i] = RiObjectBegin();
	saveRibGeometry(gdp);
	RiObjectEnd();
    }
}

void
IFD_SimpleRib::outputLights(IFD_Scene &scene)
{
    int			i;
    IFD_LightParms	parms;
    UT_String		shader;

    for (i = 0; i < myLightCount; i++)
    {
	float	color[3];

	scene.getLightParms(i, parms, myStartTime, RenderManToken);
	color[0] = parms.r;
	color[1] = parms.g;
	color[2] = parms.b;
	if (parms.isAmbient)
	{
	    myLightHandles[i] = RiLightSource("ambientlight",
						"lightcolor", color, RI_NULL);
	}
	else
	{
	    UT_DMatrix4	xform, lxform;
	    UT_Matrix4	fxform;		// Single precision matrix

	    // We have to do some wierd transform stuff to get the light
	    // transforms correct.
	    scene.getLightTransform(i, lxform, myStartTime);
	    xform.identity();
	    xform.scale(1, 1, -1);		// Transform to correct space
	    xform *= lxform;
	    fxform = xform;
	    RiTransform(fxform.matx);

	    // TODO: Real output of the shader is not easy.
	    //scene.getLightShader(i, shader, myStartTime, RenderManToken);
	    myLightHandles[i] = RiLightSource("distantlight",
						"lightcolor", color, RI_NULL);
	}
    }
}

void
IFD_SimpleRib::outputObjects(IFD_Scene &scene)
{
    int				 i, j;
    int				 gidx;
    UT_String			 name;
    const char			*name_str;
    UT_DMatrix4			 xform;
    UT_Matrix4			 fxform;
    IFD_ObjectParms		 parms;
    IFD_GeometryParms		 gparms;
    float			 color[3];
    float			 opacity[3];

    for (i = 0; i < myObjectCount; i++)
    {
	gidx = scene.getObjectGeometry(i);
	scene.getObjectName(i, name, 1);
	scene.getObjectParms(i, parms, myStartTime, RenderManToken);
	scene.getObjectTransform(i, xform, myStartTime);
	name_str = (const char *)name;

	RiAttributeBegin();
	RiAttribute("identifier", "name", &name_str, RI_NULL);
	if (parms.isMatte)
	    RiMatte(1);
	else
	{
	    for (j = 0; j < myLightCount; j++)
	    {
		if (scene.isObjectLit(i, j, myStartTime))
		     RiIlluminate(myLightHandles[j], 1);
		else RiIlluminate(myLightHandles[j], 0);
	    }

	    // TODO: Outputting the real shader isn't that easy.
	    // TODO: If we wanted, we could get the object shader transform,
	    //		and put a transform block around the shader itself.
	    // scene.getObjectShaderTransform(i, sxform, myStartTime);

	    color[0] = color[1] = color[2] = 1;
	    opacity[0] = opacity[1] = opacity[2] = 1;

	    RiColor(color);
	    RiOpacity(opacity);
	    RiSurface("plastic", RI_NULL);
	}

	fxform = xform;

	RiShadingRate(parms.shaderRate);
	RiTransform(fxform.matx);

	scene.getGeometryParms(gidx, gparms, myStartTime, RenderManToken);

	// Here, you could of course, output the geometry instead of instancing
	// (i.e. if you needed motion blur geometry).

	// RIB requires the sides to be declared not only when the geometry is
	// declared, but also when it is referenced.
	if (gparms.backfaceRemove)
	    RiSides(1);
	RiObjectInstance(myGeometryHandles[gidx]);
	if (gparms.backfaceRemove)
	    RiSides(2);		// Restore the state

	RiAttributeEnd();
    }
}

void
IFD_SimpleRib::writeTrailer()
{
    RiWorldEnd();	// End the world
    RiFrameEnd();	// End the frame
}

// --------------------------------------------------------------------
//       Here's an example of how to save some RIB geometry
// --------------------------------------------------------------------

static void
buildPolyAttribute(const GEO_PrimPoly *poly, float *array,
		    int vectorsize, int offset)
{
    int			 i, j, nvtx;
    const GEO_Point	*ppt;
    const float		*data;

    nvtx = poly->getVertexCount();
    for (i = 0; i < nvtx; i++)
    {
	ppt = poly->getVertex(i).getPt();
	if (offset < 0)		// Here we want to get position
	     data = (const float *)&ppt->getPos();
	else data = (const float *)ppt->getAttribData(offset);

	// Copy our attribute data into a vector
	for (j = 0; j < vectorsize; j++)
	    *array++ = *data++;
    }
}

static void
buildNURBSAttribute(const GEO_TPSurf *surf, float *array,
		    int vectorsize, int offset,
		    int rows, int cols, int nr, int nc)
{
    int			 r, r1, c, c1, j;
    const GEO_Point	*ppt;
    const float		*data;

    // This code assumes that the attribute being built is per-vertex (not
    // varying or uniform).  This is not the case with "Cs", so this is an
    // outstanding bug.
    for (r = 0; r < nr; r++)
    {
	r1 = r % rows;
	for (c = 0; c < nc; c++)
	{
	    c1 = c % cols;
	    ppt = surf->getVertex(r1, c1).getPt();

	    if (offset < 0)
		 data = (const float *)&ppt->getPos();
	    else data = (const float *)ppt->getAttribData(offset);
	    for (j = 0; j < vectorsize; j++)
		*array++ = *data++;
	}
    }
}

static void
outputPolygon(const GEO_PrimPoly *poly, int CsOffset)
{
    float	*P, *Cs;
    RtInt	 nverts[2];

    nverts[0] = poly->getVertexCount();
    P = new float [nverts[0] * 3];
    buildPolyAttribute(poly, P, 3, -1);
    if (CsOffset >= 0)
    {
	Cs = new float [nverts[0] * 3];
	buildPolyAttribute(poly, Cs, 3, CsOffset);

	RiGeneralPolygon(1, nverts, "P", P, "Cs", Cs, RI_NULL);
	delete [] Cs;
    }
    else
    {
	RiGeneralPolygon(1, nverts, "P", P, RI_NULL);
    }
    delete [] P;
}

static void
outputNURBS(const GEO_PrimNURBSurf *nurb, int CsOffset)
{
    GB_NUBBasis		*ub, *vb;
    float		*uknot, *vknot;
    int			 uorder, vorder;
    int			 nr, nc;
    int			 rows, cols;
    float		*Pw, *Cs;

    rows = nurb->getNumRows();
    cols = nurb->getNumCols();
    ub = (GB_NUBBasis *)nurb->getUBasis();
    vb = (GB_NUBBasis *)nurb->getVBasis();
    uorder = ub->getOrder();
    vorder = vb->getOrder();
    uknot = ub->getData();
    vknot = vb->getData();

    nr = rows;
    nc = cols;
    if (nurb->isWrappedV())
    {
	if (vb->interpolatesEnds())	nr++;
	else				nr += vorder-1;
    }
    if (nurb->isWrappedU())
    {
	if (ub->interpolatesEnds())	nc++;
	else				nc += uorder-1;
    }

    Pw = new float [nr * nc * sizeof(float)*4];
    buildNURBSAttribute(nurb, Pw, 4, -1, rows, cols, nr, nc);
    if (CsOffset >= 0)
    {
	Cs = new float [nr*nc * sizeof(float)*3];
	buildNURBSAttribute(nurb, Cs, 3, CsOffset, rows, cols, nr, nc);
    }
    else Cs = 0;

    // Now, we output the primitive
    if (Cs)
    {
	RiNuPatch(nc, uorder, uknot, uknot[uorder-1], uknot[ub->getDimension()],
		  nr, vorder, vknot, vknot[vorder-1], vknot[vb->getDimension()],
		  "Pw", Pw, "Cs", Cs, RI_NULL);
    }
    else
    {
	RiNuPatch(nc, uorder, uknot, uknot[uorder-1], uknot[ub->getDimension()],
		  nr, vorder, vknot, vknot[vorder-1], vknot[vb->getDimension()],
		  "Pw", Pw, RI_NULL);
    }
    delete [] Pw;
    delete [] Cs;
}

void
IFD_SimpleRib::saveRibGeometry(const GU_Detail *gdp)
{
    int			 i, nprim;
    int			 CsOffset;
    const GEO_Primitive	*prim;

    // Find out whether there's a color attribute on this geometry
    CsOffset = gdp->findDiffuseAttribute(GEO_POINT_DICT);

    nprim = gdp->primitives().entries();
    for (i = 0; i < nprim; i++)
    {
	prim = gdp->primitives()(i);
	switch (prim->getPrimitiveId())
	{
	    case GEOPRIMPOLY:
		outputPolygon((const GEO_PrimPoly *)prim, CsOffset);
		break;

	    case GEOPRIMNURBSURF:
		outputNURBS((const GEO_PrimNURBSurf *)prim, CsOffset);
		break;
	    // TODO:  Add the rest of the primitive types.
	}
    }
}


//
//  Now, we define the here is where we wire in the interface
//	to Houdini.

static IFD_Output *construct() { return new IFD_SimpleRib; }

void
IFDnewRenderer(void *)
{
    IFD_Entry	*custom;

    custom = new IFD_Entry(renderName, "RIB", ".tif", construct,
			    0, 0, 0, 0);
    custom->insertIntoList();	// Make sure to add it to the list
}
